/*
 *  upkg: tool for manipulating Unreal Tournament packages.
 *  Copyright (C) 2009 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include <glib-object.h>

#include "upkg.h"
#include <uobject/uobject.h>
#include <uobject/exportable.h>
#include <uobject/loadable.h>
#include <uobject/module.h>
#include <uobject/package.h>

enum {
	MODE_INFO,
	MODE_EXPORT,
	MODE_MAX
};

int verbose = 0;

static const char *progname = "upkg";
static const char *sopts = "ixvVH";
static const struct option lopts[] = {
	{ "info",     0, NULL, 'i' },
	{ "export",   0, NULL, 'x' },
	{ "verbose",  0, NULL, 'v' },
	{ "version",  0, NULL, 'V' },
	{ "help",     0, NULL, 'H' },
	{ 0 }
};

void print_version(void)
{
	printf("%s\n", PACKAGE_STRING);
	puts("\
Copyright (C) 2009 Nick Bowler.\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law."
	);
}

void print_usage(FILE *f)
{
	fprintf(f, "Usage: %s [options] package\n", progname);
}

void print_help(void)
{
	print_usage(stdout);
	puts("Detailed help coming soon.  Until then, I'll just list my options.");
	for (unsigned i = 0; lopts[i].name; i++) {
		const struct option *o = &lopts[i];
		printf("\t--%s", o->name);

		if (o->has_arg == 1) {
			printf("=val");
		} else if (o->has_arg == 2) {
			printf("[=val]");
		}

		printf(" (-%c)\n", o->val);
	}
}

static void print_upkg_flags(const char *prefix, unsigned long flags)
{
	if (flags & UPKG_FLAG_ALLOW_DOWNLOAD)
		printf("%sAllowDownload\n", prefix);
	if (flags & UPKG_FLAG_CLIENT_OPTIONAL)
		printf("%sClientOptional\n", prefix);
	if (flags & UPKG_FLAG_SERVER_ONLY)
		printf("%sServerOnly\n", prefix);
	if (flags & UPKG_FLAG_BROKEN_LINKS)
		printf("%sBrokenLinks\n", prefix);
	if (flags & UPKG_FLAG_INSECURE)
		printf("%sInsecure\n", prefix);
	if (flags & UPKG_FLAG_REQUIRED)
		printf("%sRequired\n", prefix);
}

static void print_upkg_object_flags(const char *prefix, unsigned long flags)
{
	if (flags & UPKG_OBJ_FLAG_TRANSACTIONAL)
		printf("%sTransactional\n", prefix);
	if (flags & UPKG_OBJ_FLAG_UNREACHABLE)
		printf("%sUnreachable\n", prefix);
	if (flags & UPKG_OBJ_FLAG_PUBLIC)
		printf("%sPublic\n", prefix);
	if (flags & UPKG_OBJ_FLAG_TAG_IMPORT)
		printf("%sImport\n", prefix);
	if (flags & UPKG_OBJ_FLAG_TAG_EXPORT)
		printf("%sTagExport\n", prefix);
	if (flags & UPKG_OBJ_FLAG_SOURCE_MODIFIED)
		printf("%sSourceModified\n", prefix);
	if (flags & UPKG_OBJ_FLAG_TAG_GARBAGE)
		printf("%sTagGarbage\n", prefix);
	if (flags & UPKG_OBJ_FLAG_NEED_LOAD)
		printf("%sNeedLoad\n", prefix);
	if (flags & UPKG_OBJ_FLAG_HIGHLIGHT_NAME)
		printf("%sHighlightName\n", prefix);
	if (flags & UPKG_OBJ_FLAG_IN_SINGULAR_FUNC)
		printf("%sInSingularFunc\n", prefix);
	if (flags & UPKG_OBJ_FLAG_SUPPRESSED)
		printf("%sSuppressed\n", prefix);
	if (flags & UPKG_OBJ_FLAG_IN_END_STATE)
		printf("%sInEndState\n", prefix);
	if (flags & UPKG_OBJ_FLAG_TRANSIENT)
		printf("%sTransient\n", prefix);
	if (flags & UPKG_OBJ_FLAG_PRELOADING)
		printf("%sPreloading\n", prefix);
	if (flags & UPKG_OBJ_FLAG_LOAD_FOR_CLIENT)
		printf("%sLoadForClient\n", prefix);
	if (flags & UPKG_OBJ_FLAG_LOAD_FOR_SERVER)
		printf("%sLoadForServer\n", prefix);
	if (flags & UPKG_OBJ_FLAG_LOAD_FOR_EDIT)
		printf("%sLoadForEdit\n", prefix);
	if (flags & UPKG_OBJ_FLAG_STANDALONE)
		printf("%sStandalone\n", prefix);
	if (flags & UPKG_OBJ_FLAG_NOT_FOR_CLIENT)
		printf("%sNotForClient\n", prefix);
	if (flags & UPKG_OBJ_FLAG_NOT_FOR_SERVER)
		printf("%sNotForServer\n", prefix);
	if (flags & UPKG_OBJ_FLAG_NOT_FOR_EDIT)
		printf("%sNotForEdit\n", prefix);
	if (flags & UPKG_OBJ_FLAG_DESTROYED)
		printf("%sDestroyed\n", prefix);
	if (flags & UPKG_OBJ_FLAG_NEED_POST_LOAD)
		printf("%sNeedPostLoad\n", prefix);
	if (flags & UPKG_OBJ_FLAG_HAS_STACK)
		printf("%sHasStack\n", prefix);
	if (flags & UPKG_OBJ_FLAG_NATIVE)
		printf("%sNative\n", prefix);
	if (flags & UPKG_OBJ_FLAG_MARKED)
		printf("%sMarked\n", prefix);
	if (flags & UPKG_OBJ_FLAG_ERROR_SHUTDOWN)
		printf("%sErrorShutdown\n", prefix);
	if (flags & UPKG_OBJ_FLAG_DEBUG_POST_LOAD)
		printf("%sDebugPostLoad\n", prefix);
	if (flags & UPKG_OBJ_FLAG_DEBUG_SERIALIZE)
		printf("%sDebugSerialize\n", prefix);
	if (flags & UPKG_OBJ_FLAG_DEBUG_DESTROY)
		printf("%sDebugDestroy\n", prefix);
}

void print_guid(unsigned char guid[static 16])
{
	for (unsigned i = 0; i < 16; i++) {
		printf("%02hhx", guid[i]);
		if (i == 15)
			putchar('\n');
		else
			putchar(' ');
	}
}

void print_upkg_exports(struct upkg *pkg)
{
	for (unsigned i = 0; i < pkg->export_count; i++) {
		const char *name, *package, *class;

		name  = upkg_export_name(pkg, i);
		class = upkg_export_class(pkg, i, &package);

		if (!name || !class)
			continue;

		printf("%u - %s (%s.%s)\n", i+1, name, package, class);
		printf("  Flags: %lx\n", upkg_export_flags(pkg, i));
		if (verbose >= 2) {
			print_upkg_object_flags("    ", upkg_export_flags(pkg, i));
		}
	}
}

int package_info(struct upkg *pkg)
{
	printf("Version: %u\n",  pkg->version);
	printf("License: %u\n",  pkg->license);
	printf("GUID: ");
	print_guid(pkg->guid);

	printf("Flags:   %lx\n", pkg->flags);
	if (verbose >= 2) {
		print_upkg_flags("        ", pkg->flags);
	}

	printf("Names:   %lu\n", pkg->name_count);
	if (verbose >= 3) {
		for (unsigned long i = 0; i < pkg->name_count; i++) {
			printf("\t%s\n", upkg_get_name(pkg, i));
		}
	}

	printf("Exports: %lu\n", pkg->export_count);
	if (verbose >= 1) print_upkg_exports(pkg);
	printf("Imports: %lu\n", pkg->import_count);

	return EXIT_SUCCESS;
}

static int export(struct upkg *pkg, GObject *obj, unsigned idx)
{
	char name[256];
	FILE *of;

	if (u_object_deserialize(obj, pkg, idx) != 0) {
		return -1;
	}

	if (U_OBJECT_IS_LOADABLE(obj) && u_object_load(obj) != 0) {
		return -1;
	}

	u_object_export_name(obj, name, sizeof name);

	printf("exporting %s to %s\n", upkg_export_name(pkg, idx), name);
	of = fopen(name, "wb");
	if (!of) {
		perror(name);
		return -1;
	}

	u_object_export(obj, of);

	if (fclose(of) != 0) {
		perror(name);
		return -1;
	}

	if (U_OBJECT_IS_LOADABLE(obj)) {
		u_object_unload(obj);
	}

	return 0;
}

int package_export(struct upkg *pkg)
{
	const char *class, *package;
	GObject *obj;
	GType type;
	int rc = EXIT_SUCCESS;

	for (unsigned i = 0; i < pkg->export_count; i++) {
		class = upkg_export_class(pkg, i, &package);
		if (!class) {
			fprintf(stderr, "error getting class information.\n");
			return EXIT_FAILURE;
		}

		type = u_object_module_get_class(package, class);
		if (!type) continue;

		obj = g_object_new(type, NULL);
		if (U_OBJECT_IS_EXPORTABLE(obj)) {
			if (export(pkg, obj, i) != 0) {
				rc = EXIT_FAILURE;
			}
		}
		g_object_unref(obj);
	}

	return rc;
}

int main(int argc, char **argv)
{
	UPkg *upkg;
	unsigned mode = MODE_INFO;
	int opt, rc = EXIT_FAILURE;

	if (argc > 0) progname = argv[0];

	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch (opt) {
		case 'i':
			mode = MODE_INFO;
			break;
		case 'x':
			mode = MODE_EXPORT;
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			print_version();
			return EXIT_SUCCESS;
		case 'H':
			print_help();
			return EXIT_SUCCESS;
		default:
			print_usage(stderr);
			return EXIT_FAILURE;
		}
	}

	if (argv[optind] == NULL) {
		print_usage(stderr);
		return EXIT_FAILURE;
	}

	if (u_object_module_init() != 0)
		return EXIT_FAILURE;

	upkg = U_PKG(u_pkg_new_by_file(argv[optind]));
	if (!upkg) {
		fprintf(stderr, "failed to open package!\n");
		return EXIT_FAILURE;
	}

	switch (mode) {
	case MODE_INFO:
		rc = package_info(upkg->pkg);
		break;
	case MODE_EXPORT:
		rc = package_export(upkg->pkg);
		break;
	}

	g_object_unref(upkg);
	u_object_module_exit();
	return rc;
}
